import time
from dataclasses import dataclass
from typing import Callable, Set, Tuple, Any

from pypsql_api.wire.actions import write_ssl_resp_no, read_startup_message, read_plain_text_password_request, \
    write_plain_text_password_request, read_ssl_request, write_ready_for_query, write_auth_ok, read_receive_command, \
    write_empty_response, write_data_frame_response
from pypsql_api.wire.actions_types import Names
from pypsql_api.wire.query.simple_query import process_simple_query


@dataclass
class State:
    allowed: Set[str]
    name: str
    fn: Callable[[Any], Tuple[Any, str]]

    def move_next(self, context: Any):
        return self.fn(context)


def move_state(context: Any, state: State) -> Tuple[Any, str]:
    ctx_new, new_state = state.move_next(context)

    if not new_state:
        return ctx_new, Names.CLOSE

    if not new_state in state.allowed:
        raise Exception(f"Invalid state {new_state} from {state.name}, allowed states are {state.allowed}")

    return ctx_new, new_state


processes = {
    Names.READ_SSL_REQUEST: State(name=Names.READ_SSL_REQUEST, allowed={Names.WRITE_SSL_NO}, fn=read_ssl_request),
    Names.WRITE_SSL_NO: State(name=Names.READ_SSL_REQUEST, allowed={Names.READ_STARTUP_MESSAGE}, fn=write_ssl_resp_no),
    Names.READ_STARTUP_MESSAGE: State(name=Names.READ_STARTUP_MESSAGE,
                                      allowed={Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST}, fn=read_startup_message),

    Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST: State(name=Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST,
                                                   allowed={Names.READ_PLAIN_TEXT_PASSWORD},
                                                   fn=write_plain_text_password_request),

    Names.READ_PLAIN_TEXT_PASSWORD: State(name=Names.READ_PLAIN_TEXT_PASSWORD,
                                          allowed={Names.CLOSE, Names.WRITE_AUTH_OK},
                                          fn=read_plain_text_password_request),

    Names.WRITE_AUTH_OK: State(name=Names.READ_PLAIN_TEXT_PASSWORD,
                               allowed={Names.READY_FOR_QUERY},
                               fn=write_auth_ok),

    Names.READY_FOR_QUERY: State(name=Names.READY_FOR_QUERY,
                                 allowed={Names.CLOSE, Names.RECEIVE_COMMAND}, fn=write_ready_for_query),

    Names.RECEIVE_COMMAND: State(name=Names.RECEIVE_COMMAND,
                                 allowed={Names.CLOSE, Names.READY_FOR_QUERY, Names.SIMPLE_QUERY},
                                 fn=read_receive_command),

    Names.SIMPLE_QUERY: State(name=Names.SIMPLE_QUERY,
                              allowed={Names.CLOSE, Names.READY_FOR_QUERY, Names.WRITE_EMPTY_RESPONSE, Names.WRITE_DATA_FRAME},
                              fn=process_simple_query),

    Names.WRITE_EMPTY_RESPONSE: State(name=Names.WRITE_EMPTY_RESPONSE,
                                      allowed={Names.CLOSE, Names.READY_FOR_QUERY},
                                      fn=write_empty_response),

    Names.WRITE_DATA_FRAME: State(name=Names.WRITE_DATA_FRAME,
                                      allowed={Names.CLOSE, Names.READY_FOR_QUERY},
                                      fn=write_data_frame_response)

}


def run_states(context, state_name):
    while state_name != Names.CLOSE:
        print(f"Running state {state_name}")
        state = processes.get(state_name)

        context, state_name = move_state(context, state)

        if not state_name:
            state_name = Names.CLOSE

    print("End run states")

    return context, state_name



import io
import time
from dataclasses import dataclass
from typing import Any, Optional, Tuple

from pypsql_api.wire.actions_types import Names
from pypsql_api.wire.bytes import ReadingIO


@dataclass
class SSLRequest:
    request_code: int

    @staticmethod
    def read(buff: ReadingIO):
        inner_buff, message_len = buff.read_int32_delim_message()
        return SSLRequest(request_code=inner_buff.read_int32())


@dataclass
class StartupMessage:
    protocol_version: int
    user: str
    database: str
    options: str

    @staticmethod
    def read(buff: ReadingIO):
        inner_buff, message_len = buff.read_int32_delim_message()

        if not inner_buff:
            raise Exception("Invalid startup message, the length must be > 4 bytes")

        protocol_version = inner_buff.read_int32()
        message_len -= 4
        d = {}

        if message_len > 0:
            while True:
                n = inner_buff.read_cstring()
                if not n:
                    break
                v = inner_buff.read_cstring()
                if not v:
                    break

                d[n] = v

        return StartupMessage(
            protocol_version=protocol_version, user=d.get('user', ''), database=d.get('database', ''),
            options=d.get('options', '')
        )


@dataclass
class PasswordMessage:
    password: str

    @staticmethod
    def read_body(buff: ReadingIO):
        return PasswordMessage(password=buff.read_cstring())


@dataclass
class QueryMessage:
    query: str

    process_name: Any = Names.SIMPLE_QUERY

    @staticmethod
    def read_body(buff: ReadingIO):
        return QueryMessage(query=buff.read_cstring())


@dataclass
class Terminate:
    process_name: Any = Names.CLOSE

    @staticmethod
    def read_body(_: ReadingIO):
        return Terminate()


message_type_map = {
    ord('p'): PasswordMessage.read_body,
    ord('Q'): QueryMessage.read_body,
    ord('X'): Terminate.read_body,
}


class Message:

    @staticmethod
    def read(buff: ReadingIO) -> Tuple[Optional[Any], Optional[bytes]]:
        """
        Returns [Optional[Message], type:bytes
        If The message and Type are None, we've reached the end of stream
        """

        t = buff.read_byte()

        if not t:
            return None, None

        fn = message_type_map.get(ord(t), None)
        if fn:
            try:
                inner_buff, _ = buff.read_int32_delim_message()
                return fn(inner_buff), t
            except Exception as e:
                return fn(io.BytesIO()), t

        return None, t



import enum
from dataclasses import dataclass
from typing import Dict, Any

from pypsql_api.config.types import Session
from pypsql_api.ext.handlers import AuthHandler, QueryHandler
from pypsql_api.wire.bytes import ReadingIO, WritingIO


@dataclass
class Context:
    input: ReadingIO
    output: WritingIO

    session: Session

    auth_handler: AuthHandler
    query_handler: QueryHandler

    mem: Dict[str, Any]  # keep memory between tasks

    def update_mem(self, k, v):
        return Context(input=self.input, output=self.output,
                       session=self.session,
                       auth_handler=self.auth_handler,
                       query_handler=self.query_handler,
                       mem={**self.mem, **{k: v}})


class Names(enum.Enum):
    ERROR = 0
    READ_SSL_REQUEST = 1
    WRITE_SSL_YES = 2
    WRITE_SSL_NO = 3

    WRITE_PLAIN_TEXT_PASSWORD_REQUEST = 4
    READ_PLAIN_TEXT_PASSWORD = 5

    READ_STARTUP_MESSAGE = 6

    CLOSE = 7

    READY_FOR_QUERY = 8
    WRITE_AUTH_OK = 9
    RECEIVE_COMMAND = 10  # after ready for query we wait for the next command
    SIMPLE_QUERY = 11

    WRITE_EMPTY_RESPONSE = 12
    WRITE_DATA_FRAME = 13



from typing import Tuple, Optional, Union, Any

from pandas import DataFrame

from pypsql_api.config.types import Session


class AuthHandler:

    def handle(self, session: Session) -> Tuple[bool, Optional[str]]:
        """
        :param session:
        :return: Tuple[bool, Optional[str]] if (true, any) then AuthOK, otherwise AuthFail, and the message in the str
        contains any error message
        """
        raise Exception("Not implemented")


class QueryHandler:

    def handle(self, session: Session, sql) -> Tuple[Optional[Union[Any, Optional[DataFrame]]], Optional[Any]]:
        """
        :param session:
        :param sql: any sql statement passed by the client,
         this can be multiple statements and being-commit statements.
        :return: None -> No result
                 (Any, Msg)  -> Msg for future use to return notification messages.
                 (DataFrame, Any) -> The DataFrame the contains the data and schema returned
        """
        raise Exception("Not implemented")
